from django.contrib.auth.models import AbstractUser
from django.db import models

from django_autoshard import utils
from django_autoshard.managers import ShardedManager, UserManager


class ShardedModel(models.Model):
    id = models.BigIntegerField(primary_key=True)

    objects = ShardedManager()

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super(ShardedModel, self).__init__(*args, **kwargs)
        try:
            key = getattr(self, self.SHARD_KEY)  # kwargs[self.SHARD_KEY]
        except AttributeError:
            raise RuntimeError('{} does not define a SHARD_KEY'.format(self.__repr__()))
        except KeyError:
            raise RuntimeError('{} does not have {} set'.format(self.__repr__(), self.SHARD_KEY))
        self.__shard = utils.get_shard(key)

    @property
    def shard(self):
        return self.__shard

    def save(self, *args, **kwargs):
        kwargs['using'] = self.shard.alias
        if self.pk is not None:
            return super(ShardedModel, self).save(*args, **kwargs)  # UPDATE
        self.id = utils.generate_uuid(self.shard.index)
        return super(ShardedModel, self).save(*args, **kwargs)  # Set the auto-id


class ShardRelatedModel(models.Model):
    class Meta:
        abstract = True

    @property
    def shard(self):
        for field in self._meta.fields:
            pass


class User(ShardedModel, AbstractUser):
    SHARD_KEY = 'email'

    objects = UserManager()



from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Build drop shards queries and print them. Will never actually run the queries. Useful in development.'

    def handle(self, *args, **opts):
        for _, shard in settings.SHARDS.items():
            self.stdout.write('DROP DATABASE %s;' % shard.database)



from collections import OrderedDict

from django.conf import settings as django_settings
from django.core import exceptions
from .shard import Shard
from . import utils


class ShardingFactory:

    def configure(self):
        shards = dict()
        nodes = django_settings.DATABASES.copy()
        for node, db in nodes.items():
            try:
                _ = db['NAME']
            except KeyError:
                raise exceptions.ImproperlyConfigured('Node {} does not have a database name.'.format(node))
            try:
                _ = db['RANGE']
            except KeyError:
                raise exceptions.ImproperlyConfigured('Node {} does not specify a shard range.'.format(node))

            node_shards = self.set_logical_shards(node, db)
            shards.update(node_shards)

        django_settings.SHARDS = OrderedDict(sorted(shards.items()))
        django_settings.DATABASE_ROUTERS = ('django_autoshard.routers.ShardRouter', )

    def set_logical_shards(self, node, config):
        result = dict()
        for i in config['RANGE']:
            shard = config.copy()
            del shard['RANGE']
            shard['NAME'] = '{}_{}'.format(config['NAME'], i)
            alias = '{}_{}'.format(node, i)
            django_settings.DATABASES[alias] = shard
            node_index = utils.get_shard_index(alias)
            replicas = self.set_replicas(config)
            result[node_index] = Shard(node_index, alias, replicas)
        return result

    def set_replicas(self, config):
        return {}



from django.contrib.auth import get_user_model
from django_autoshard.tests.base import TestBase

User = get_user_model()


class UserTests(TestBase):
    def test__create__user__success(self):
        u1 = User.objects.create(
            username='test',
            email='test@local.dev'
        )
        u2 = User.objects.get(email='test@local.dev')
        self.assertEqual(u1, u2)



from django.core.management import call_command
from django.test.runner import DiscoverRunner


class TestRunner(DiscoverRunner):
    def setup_databases(self, **kwargs):
        super(TestRunner, self).setup_databases(**kwargs)
        call_command('migrate_shards')

    def run_tests(self, test_labels=None, extra_tests=None, **kwargs):
        result = super(TestRunner, self).run_tests(['django_autoshard.tests'], extra_tests=None, **kwargs)
        return result

    def teardown_databases(self, old_config, **kwargs):
        try:
            super(TestRunner, self).teardown_databases(old_config, **kwargs)
        except TypeError:
            pass



from itertools import chain
import time
import six

if six.PY3:
    from concurrent.futures import ThreadPoolExecutor
else:
    from collections import deque
    from threading import Thread

from django.conf import settings
from django.core.management.commands.migrate import Command as DjangoMigrateCommand
from django.db import OperationalError


class Command(DjangoMigrateCommand):
    def migrate(self, shard, *args, **options):
        try:
            start = time.time()
            options['database'] = shard.alias
            super(Command, self).handle(*args, **options)
            shard.connection.close()
            self.stdout.write(
                self.style.MIGRATE_HEADING('Migrated database %s in %.4f sec' % (shard.database, time.time() - start)))
        except OperationalError as e:
            self.stdout.write(
                    self.style.MIGRATE_HEADING('Failed migrating database: %s. Error: %s' % (options['database'], e)))

    def run(self, *args, **options):
        if six.PY3:
            executor = ThreadPoolExecutor(max_workers=10)
            for _, shard in settings.SHARDS.items():
                executor.submit(self.migrate, *chain((shard,), args), **options)
        else:
            queue = deque()
            for _, shard in settings.SHARDS.items():
                queue.append(Thread(target=self.migrate, args=chain((shard,), args), kwargs=options))
            workers = 0
            while queue:
                if workers < 10:
                    t = queue.popleft()
                    t.start()

    def handle(self, *args, **options):
        options['verbosity'] = 0
        try:
            self.run(*args, **options)
            self.stdout.write(self.style.MIGRATE_HEADING('Done.\n'))
        except KeyboardInterrupt:
            self.stdout.write(self.style.MIGRATE_HEADING('Migration interrupted.\n'))



DEBUG = False
SECRET_KEY = '0mgs0s3cr37!10n3!!'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'autoshard',
        'RANGE': range(2),
    }
}
MIDDLEWARE_CLASSES = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'django_autoshard'
]
AUTH_USER_MODEL = 'django_autoshard.User'
TEST_RUNNER = 'django_autoshard.tests.runner.TestRunner'



from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.db import connection
from django.db.utils import OperationalError, ProgrammingError
from django_autoshard import models


class Command(BaseCommand):
    help = 'Create shards databases'

    def add_arguments(self, parser):
        parser.add_argument('--list', '-l', action='store_true', default=False,
                            help='Show a list of all the constraints that will be dropped')

    def handle(self, *args, **options):
        with connection.cursor() as cursor:
            for model in apps.get_models():
                if issubclass(model, models.ShardedModel) or issubclass(model, models.ShardRelatedModel):
                    continue
                self.run(cursor, model, **options)

    def run(self, cursor, model, **options):
        try:
            constraints = dict()
            for key, val in cursor.db.introspection.get_constraints(cursor, model._meta.db_table).items():
                if val['foreign_key'] is not None and val['foreign_key'][0] == get_user_model()._meta.db_table:
                    constraints[key] = val
        except ProgrammingError:
            return

        if len(constraints) == 0:
            f = self.style.MIGRATE_HEADING('No constraints defined for {}.'.format(str(model)))
            self.stdout.write(f)
            return

        for key, val in constraints.items():
            db_table, _ = val['foreign_key']

            sql = cursor.db.SchemaEditorClass.sql_delete_fk % dict(
                table=model._meta.db_table,
                name=key
            )
            if options.get('list'):
                f = self.style.MIGRATE_HEADING('{}'.format(sql))
                self.stdout.write(f)
                continue

            try:
                f = self.style.MIGRATE_HEADING('Executing {}'.format(sql))
                self.stdout.write(f)

                cursor.execute(sql)
                f = self.style.MIGRATE_HEADING('Done.\n')
                self.stdout.write(f)
            except OperationalError as e:
                f = self.style.MIGRATE_HEADING('Failed [{}].\n'.format(str(e)))
                self.stdout.write(f)



from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'List all installed shards'

    def handle(self, *args, **opts):
        for _, shard in settings.SHARDS.items():
            info = 'ALIAS [{}] :: DATABASE NAME [{}] :: HOST :: [{}]'.format(
                shard.alias, shard.database, shard.host
            )
            self.stdout.write(self.style.MIGRATE_HEADING(info))



from django_autoshard.models import ShardedModel, ShardRelatedModel


class ShardRouter:
    def db_for_write(self, model, **hints):
        if not issubclass(model, ShardedModel) and not issubclass(model, ShardRelatedModel):
            return 'default'
